/*==============================================================================

    S E N S O R S   AMBIENT LIGHT   D R I V E R

DESCRIPTION

   Implements the driver for the al3320 ALS(Light) Sensor

Copyright (c) 2015 by QUALCOMM Technologies, Incorporated.  All Rights Reserved.
Qualcomm Technologies Proprietary and Confidential
==============================================================================*/
/*==============================================================================

                      EDIT HISTORY FOR FILE

  This section contains comments describing changes made to the module.
  Notice that changes are listed in reverse chronological order.

$Header: //source/qcom/qct/core/pkg/dsps/rel/2.2/adsp_proc/Sensors/dd/qcom/src/sns_dd_als_al3320_priv.h#1 $

***************************************************************************
* REVISON HISTORY
* VERSION   DATE(MM/DD/YY)  WHO     DESCRIPTION
* 2.0       03/20/2017      LO      Refactoring scenario of als auto gain
* 1.0       12/20/2016      LO      Initial version of driver for AL3320
***************************************************************************
==============================================================================*/

#ifndef _SNSAL3320_H
#define _SNSAL3320_H

#include "fixed_point.h"
#include "sns_ddf_util.h"
#include <stdbool.h>
#include <stdio.h>
#include "sns_ddf_attrib.h"
#include "sns_ddf_common.h"
#include "sns_ddf_comm.h"
#include "sns_ddf_driver_if.h"
#include "sns_ddf_smgr_if.h"
#include "sns_ddf_util.h"
#include "sns_ddf_memhandler.h"
#include "sns_ddf_signal.h"

#define AL3320_ENABLE_UIMG_SUPPORT 0
#if AL3320_ENABLE_UIMG_SUPPORT
#else
#define sns_ddf_malloc_ex(ptr, size, smgr_handle)                               sns_ddf_malloc(ptr, size)
#define sns_ddf_mfree_ex(ptr, smgr_handle)                                      sns_ddf_mfree(ptr)
#define sns_ddf_memhandler_malloc_ex(memhandler, size, smgr_handle)             sns_ddf_memhandler_malloc(memhandler, size)
#define sns_ddf_smgr_is_in_uimage()     false
#endif

sns_ddf_driver_if_s sns_alsprx_al3320_driver_fn_list;
#define AL3320_DD_IF sns_alsprx_al3320_driver_fn_list

#define REG_FIXED_NV_SIZE_128            128
#define LSC_MAGIC_NUM                    50
#define LSC_DRIVER_ID                    1
#define INVAILD_GPIO                     0xffff
#define AL3320_ALS_AUTO_GAIN_EN          0
#define AL3320_ALS_AUTO_GAIN_USDELAY     200
#define SNS_ALS_AUTO_GAIN_MAX_DATA       65000
#define SNS_ALS_AUTO_GAIN_HTHRES         ((3*SNS_ALS_AUTO_GAIN_MAX_DATA)/4)
#define SNS_ALS_AUTO_GAIN_LTHRES         ((2*SNS_ALS_AUTO_GAIN_MAX_DATA)/16)


#define AL3320_STA_ALS_INTR              0x01
#define AL3320_CLS_ALS_INTR              0x00
#define AL3320x_ENABLE_MSK_AEN           0x01
#define AL3320x_ALSPRX_PWR_DOWN          0x00
#define AL3320_ALS_PERSIS_VAL            0x03
#define SNS_DD_ALSPRX_NUM_SENSOR_TYPES   2
#define SNS_DD_ALS_PWR                   170
#define SNS_DD_ALS_RES                   FX_FLTTOFIX_Q16(1.0)
#define SNS_DD_ALSPRX_LO_PWR             2
#define SNS_DD_ALS_ACCURACY_DIV_FACTOR   4
#define SNS_DD_ALS_HALF_LUX              500
#define SNS_DD_ALSPRX_PRX_BITS           16
#define SNS_DD_ALSPRX_PRX_FREQ           80
#define SNS_DD_ALSPRX_ALS_BITS           16
#define SNS_DD_ALSPRX_ALS_FREQ           10


#define SNS_CLR_ALL_INT                  0x00
#define SNS_DD_ALS_PWR_DOWN              0x00
#define SNS_DD_ALS_INT_CTRL_POLL         0x00
#define SNS_DD_ALS_INT_CTRL_DRI          0x08
#define SNS_DD_ALSPRX_WAITE              0x04
#define SNS_ALS_F_ODR                    0
#define SNS_ALSPRO_INT_ODR               10

#define SNS_ALS_INT_CNT_L                10
#define SNS_ALS_INT_CNT_H                0xfff5
#define SNS_ALS_THR_LOW_F                0xffff
#define SNS_ALS_THR_HIG_F                0x0000
#define SNS_ALS_CAL_TARGET               0x40
#define SNS_RANGE_MASK                   0x06

#define AL3320_ALS_GAIN_33               0x00
#define AL3320_ALS_GAIN_08               0x01
#define AL3320_ALS_GAIN_02               0x02
#define AL3320_ALS_GAIN_01               0x03
#define AL3320_ALS_GAIN_SHIFT            1
#define AL3320_ALS_GAIN_MASK             0x06
#define AL3320_ALS_GAIN_33_RATION        0.512
#define AL3320_ALS_GAIN_08_RATION        0.128
#define AL3320_ALS_GAIN_02_RATION        0.032
#define AL3320_ALS_GAIN_01_RATION        0.01


//system registers
#define AL3320_MODE_COMMAND              0x00
#define AL3320_FLAG_STATUS               0x01
#define AL3320_INT_CLEAR                 0x01
#define AL3320_INT_CTRL                  0x02
#define AL3320_WAITE_TIME                0x06
#define AL3320_INT_CLEAR_MASK            0x7c
#define AL3320_ALS_CH0_ADC_LSB           0x2c
#define AL3320_ALS_CH0_ADC_MSB           0x2d
#define AL3320_ALS_CH1_ADC_LSB           0x2e
#define AL3320_ALS_CH1_ADC_MSB           0x2f
#define AL3320_PX_ADC_LSB                0x26
#define AL3320_PX_ADC_MSB                0x27

//als registers
#define AL3320_ALS_CON                   0x07
#define AL3320_ALS_PERSIS                0x08
#define AL3320_ALS_TIME                  0x0a
#define AL3320_ALS_LDATA_L               0x22
#define AL3320_ALS_LDATA_H               0x23
#define AL3320_ALS_LTHL                  0x30
#define AL3320_ALS_LTHH                  0x31
#define AL3320_ALS_HTHL                  0x32
#define AL3320_ALS_HTHH                  0x33

typedef enum {
    IRQ_TYPE,
    TIMER_TYPE
} AL3320_SIG_TYPE;

typedef union {
    uint16_t   word;
    uint8_t    byte[2];
} uword_u;

typedef struct  {
    uint16_t data;
    uint32_t lux;
    uint32_t mlux;
    bool     lux_ok;
} als_info;

/* proximity data */
typedef struct  {
    uint16_t data;
    int16_t  event;
} prox_info;

typedef struct {
    uint8_t  gain;
    uint8_t  persis;
    uint16_t cal_target;
    uint16_t thresh_low;
    uint16_t thresh_high;
    uint32_t odr;
    bool     enabled;
} als_setup;

typedef struct {
    uint8_t  pgain;
    uint8_t  led_pulse;
    uint8_t  led_ratio;
    uint16_t thresh_low;
    uint16_t thresh_high;
    uint32_t odr;
    bool     enabled;
    uint8_t  offset_l;
    uint8_t  offset_h;
} prox_setup;

typedef struct {
    uint8_t  enable;
    uint8_t  config;
    uint8_t  control;
    uint8_t  int_ctrl;
    uint8_t  flag;
} al3320_reg;

typedef struct  {
    als_setup   als;
    prox_setup  prox;
    al3320_reg  reg;
} lsc_setup;

typedef struct {
    als_info      als;    
    prox_info     prox;
} lsc_cur_info;

typedef struct {
    lsc_cur_info       data;
    lsc_setup          setup;
    uint32_t           int_gpio;
} al3320_chip;

typedef enum
{
    SNS_DD_AL3320_SUCCESS = 0,
    SNS_DD_AL3320_ALS_CAL_FAILED_LUX = 1, /*ALS CAL error - Received lux value out of range*/
    SNS_DD_AL3320_ALS_CAL_FAILED_RST = 2, /*ALS CAL error - cannot reset the device*/
    SNS_DD_AL3320_ALS_CAL_FAILED_REG = 3, /*ALS CAL error - cannot write back to registry*/
    SNS_DD_AL3320_ALS_CAL_FAILED_ENB = 4,  /*ALS CAL error - cannot enable (error in set power state)*/
    SNS_DD_AL3320_PROX_CAL_FAILED_CHIP = 5, /*PROX CAL error - cannot read/write the device*/
    SNS_DD_AL3320_PROX_CAL_FAILED_RST  = 6, /*PROX CAL error - cannot reset the device*/
    SNS_DD_AL3320_PROX_CAL_FAILED_REG  = 7, /*PROX CAL error - cannot write back to registry*/
    SNS_DD_AL3320_PROX_CAL_FAILED_ENB  = 8,  /*PROX CAL error - cannot enable (error in set power state)*/
    SNS_DD_AL3320_PROX_CAL_FAILED_DATA = 9  /*PROX CAL error - calibration data is too big*/
} sns_dd_test_err_e;

typedef enum
{
  SNS_PRX_FAR_AWAY,
  SNS_PRX_NEAR_BY,
  SNS_PRX_NEAR_BY_UNKNOWN
} sns_dd_prx_nearby_type;

typedef  enum
{
  SNS_DD_DEV_STOPPED,                         /* 0 */
  SNS_DD_DEV_CONFIGURED,                      /* 1 */
  SNS_DD_DEV_GET_DATA,                        /* 2 */
  SNS_DD_DEV_MAX_STATES                       /* 3 */
}sns_dd_device_state_type;

typedef enum
{
  SNS_DD_ALSPRX_NV_SOURCE_DEFAULT,
  SNS_DD_ALSPRX_NV_SOURCE_REG
}sns_dd_nv_source;

/* data structure for light(ALS) driver */
typedef struct
{
  uint32_t                  last_mlux;
  uint16_t                  accuracy;
} sns_dd_als_db_type;

/* data structure for NV items */
typedef struct
{
  uint8_t  visible_ratio;   /* visible light transmission ratio of the glass/pipe in % */
  uint8_t  ir_ratio;        /* IR transmission ratio of light glass/pipe in % */
  uint16_t dc_offset;       /* DC offset in ADC count */
  uint16_t thresh_near;     /* near detection threshold in ADC count adjusted by DC offset */
  uint16_t thresh_far;      /* far detection threshold in ADC count adjusted by DC offset */
  uint16_t prx_factor;      /* PRX multiplicative factor in % */
  uint16_t als_factor;      /* ALS multiplicative factor in % */
  /* Device/Driver Specific data fields - start*/
  uint32_t lsc_num;         /* unique number for LSM */
  uint32_t drv_ID;          /* identify specific driver */
  uint32_t calibratePhone;  /* if true, calibrate phone (should only have to done during manufacture) */
  uint32_t device;          /* current sensor */
  uint32_t ALS_thresh_high; /* ALS Interrupt Threshold High in % */
  uint32_t ALS_thresh_low;   /* ALS Interrupt Threshold Low in % */
} sns_dd_nv_db_type;

typedef struct
{
  sns_dd_nv_db_type         nv_db;                  /* NV item DB type */
  sns_dd_nv_source          nv_source;
  sns_ddf_powerstate_e      pwr_mode;
  sns_dd_device_state_type  state;
} sns_dd_als_common_db_type;

/* State struct for ALS/PRX */
typedef struct {
   sns_ddf_handle_t               smgr_handle;    /* SDDI handle used to notify_data */
   sns_ddf_handle_t               port_handle;    /* handle used to access the I2C bus */
   uint32_t                       dropped_logs;
   sns_dd_als_db_type             sns_dd_als_db;
   sns_dd_als_common_db_type      sns_dd_als_common_db;
   bool                           calibration_mode;    /* calibration mode enable flag */
   uint16_t                       nv_db_size;   /* size of NV item */
   al3320_chip                    chip;
   sns_ddf_handle_t               ZBW_port_handle;
   uint32_t                       int_gpio;
   sns_ddf_sensor_data_s*         s_data;
   sns_ddf_sensor_sample_s*       samples;
   bool                           als_auto_gain_trig;
   sns_ddf_time_t                 als_auto_gain_trig_time;
   uint32_t                       als_auto_gain_delay_time;
} sns_dd_als_state_t;


sns_ddf_status_e sns_dd_als_al3320_init
(
  sns_ddf_handle_t*        dd_handle_ptr,
  sns_ddf_handle_t         smgr_handle,
  sns_ddf_nv_params_s*     nv_params,
  sns_ddf_device_access_s  device_info[],
  uint32_t                 num_devices,
  sns_ddf_memhandler_s*    memhandler,
  sns_ddf_sensor_e*        sensors[],
  uint32_t*                num_sensors
);

sns_ddf_status_e sns_dd_als_al3320_get_data
(
  sns_ddf_handle_t        dd_handle,
  sns_ddf_sensor_e        sensors[],
  uint32_t                num_sensors,
  sns_ddf_memhandler_s*   memhandler,
  sns_ddf_sensor_data_s*  data[] /* ignored by this async driver */
);

sns_ddf_status_e sns_dd_als_al3320_set_attrib
(
  sns_ddf_handle_t     dd_handle,
  sns_ddf_sensor_e     sensor,
  sns_ddf_attribute_e  attrib,
  void*                value
);

sns_ddf_status_e sns_dd_als_al3320_get_attrib
(
  sns_ddf_handle_t     dd_handle,
  sns_ddf_sensor_e     sensor,
  sns_ddf_attribute_e  attrib,
  sns_ddf_memhandler_s* memhandler,
  void**               value,
  uint32_t*            num_elems
);

void sns_dd_als_al3320_handle_irq
(
  sns_ddf_handle_t dd_handle,
  uint32_t          gpio_num,
  sns_ddf_time_t    timestamp
);

sns_ddf_status_e sns_dd_als_al3320_reset
(
  sns_ddf_handle_t dd_handle
);

sns_ddf_status_e sns_dd_als_al3320_run_test
(
 sns_ddf_handle_t  dd_handle,
 sns_ddf_sensor_e  sensor,
 sns_ddf_test_e    test,
 uint32_t*         err
);
sns_ddf_status_e sns_dd_als_al3320_enable_sched_data
(
 sns_ddf_handle_t state,
 sns_ddf_sensor_e sensor,
 bool             enable
 );
 
sns_ddf_status_e sns_dd_als_al3320_probe(
  sns_ddf_device_access_s* device_info,
  sns_ddf_memhandler_s*    memhandler,
  uint32_t*                num_sensors,
  sns_ddf_sensor_e**       sensors );

sns_ddf_status_e sns_dd_als_al3320_sensor_samples(sns_dd_als_state_t   *state,
                                            sns_ddf_sensor_data_s   *data_ptr,
                                            sns_ddf_memhandler_s    *memhandler,
                                            sns_ddf_sensor_sample_s *sample_data);

sns_ddf_status_e sns_dd_als_al3320_update_registers(sns_dd_als_state_t* state);

sns_ddf_status_e sns_dd_als_al3320_get_mlux(sns_dd_als_state_t* state);

sns_ddf_status_e sns_dd_als_al3320_get_als_data(sns_dd_als_state_t* state);

sns_ddf_status_e sns_dd_als_al3320_write_byte(sns_ddf_handle_t port_handle, uint8_t reg, uint8_t* byte_val);

sns_ddf_status_e sns_dd_als_al3320_write_word(sns_ddf_handle_t port_handle, uint8_t reg, uword_u* word_val);

sns_ddf_status_e sns_dd_als_al3320_read_byte(sns_ddf_handle_t port_handle, uint8_t reg, uint8_t* byte_val);

sns_ddf_status_e sns_dd_als_al3320_read_word(sns_ddf_handle_t port_handle, uint8_t reg, uword_u* word_val); 

bool sns_dd_al3320_als_auto_gain(sns_dd_als_state_t* state, uint16_t alsdata);

#endif
