/*==============================================================================

    S E N S O R S   AMBIENT LIGHT   D R I V E R

DESCRIPTION

   Implements the driver for the al3320 ALS(Light) Sensor

Copyright (c) 2015 by QUALCOMM Technologies, Incorporated.  All Rights Reserved.
Qualcomm Technologies Proprietary and Confidential
==============================================================================*/
/*==============================================================================

                      EDIT HISTORY FOR FILE

  This section contains comments describing changes made to the module.
  Notice that changes are listed in reverse chronological order.

$Header: //source/qcom/qct/core/pkg/dsps/rel/2.2/adsp_proc/Sensors/dd/qcom/src/sns_dd_als_al3320_uimg.c#1 $


***************************************************************************
* REVISON HISTORY
* VERSION   DATE(MM/DD/YY)  WHO     DESCRIPTION
* 2.0       03/20/2017      LO      Refactoring scenario of als auto gain
* 1.0       12/20/2016      LO      Initial version of driver for AL3320
***************************************************************************
==============================================================================*/
/*============================================================================
                                INCLUDE FILES
============================================================================*/
#include "sns_dd_als_al3320_priv.h"
        
#include "sns_memmgr.h"
#include "log_codes.h"
#include "sns_debug_str.h"
#include "sns_debug_api.h"
#include "sns_log_types.h"
#include "sns_log_api.h"


/*===========================================================================
FUNCTION      sns_dd_als_al3320_write_byte

DESCRIPTION   write a byte from device
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_write_byte(sns_ddf_handle_t port_handle, uint8_t reg, uint8_t* byte_val)
{
    sns_ddf_status_e status;
    uint8_t          write_count;

    status = sns_ddf_write_port(port_handle,
                                reg,
                                byte_val,
                                1,
                                &write_count
                                );                              
    return status;
}
/*===========================================================================
FUNCTION      sns_dd_als_al3320_write_word

DESCRIPTION   write a word from device
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_write_word(sns_ddf_handle_t port_handle, uint8_t reg, uword_u* word_val)
{
    sns_ddf_status_e status;
    uint8_t          write_count;

    status = sns_ddf_write_port(port_handle,
                                reg,
                                &word_val->byte[0],
                                2,
                                &write_count
                                );                              
    return status;
}
/*===========================================================================
FUNCTION      sns_dd_als_al3320_read_byte

DESCRIPTION   read a byte from device
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_read_byte(sns_ddf_handle_t port_handle, uint8_t reg, uint8_t* byte_val)
{
    sns_ddf_status_e status;
    uint8_t          read_count;

    status = sns_ddf_read_port(port_handle,
                               reg,
                               byte_val,
                               1,
                               &read_count
                               );                              
    return status;
}
/*===========================================================================
FUNCTION      sns_dd_als_al3320_read_word

DESCRIPTION   read a word from device
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_read_word(sns_ddf_handle_t port_handle, uint8_t reg, uword_u* word_val)
{
    sns_ddf_status_e status;
    uint8_t          read_count;
    uword_u          word_tmp;

    status = sns_ddf_read_port(port_handle,
                               reg,
                               &word_tmp.byte[0],
                               2,
                               &read_count
                               );
    if(status == SNS_DDF_SUCCESS)
    {
        word_val->word = word_tmp.word;
    }
    return status;
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_log

DESCRIPTION   Log the latest sensor data
===========================================================================*/
void sns_dd_als_al3320_log(
   sns_dd_als_state_t* dd_ptr,
   sns_ddf_sensor_e       sensor_type,
   uint32_t               data1,
   uint32_t               data1_q16,
   uint32_t               data2,
   uint32_t               data2_q16,
   uint32_t               raw_data )
{
   sns_err_code_e err_code;
   sns_log_sensor_data_pkt_s *log_struct_ptr;
   log_pkt_t log_pkt_type;
    
#ifndef AL3320_DDK
   log_pkt_type = SNS_LOG_CONVERTED_SENSOR_DATA;
#else
   log_pkt_type = SNS_LOG_CONVERTED_SENSOR_DATA_PUBLIC;
#endif

   DDF_MSG_0(HIGH, "AL3320 Log Data, start");

   /* Allocate log packet */
   err_code = sns_logpkt_malloc(log_pkt_type,
                                sizeof(sns_log_sensor_data_pkt_s),
                                (void**)&log_struct_ptr);

   if ((err_code == SNS_SUCCESS) && (log_struct_ptr != NULL))
   {
     log_struct_ptr->version = SNS_LOG_SENSOR_DATA_PKT_VERSION;
     log_struct_ptr->sensor_id = sensor_type;
     log_struct_ptr->vendor_id = SNS_DDF_VENDOR_AMS;

     /* Timestamp the log with sample time */
     log_struct_ptr->timestamp = sns_ddf_get_timestamp();

     /* Log the sensor data */
     if (sensor_type == SNS_DDF_SENSOR_AMBIENT)
     {
       log_struct_ptr->num_data_types = 3;
       log_struct_ptr->data[0] = data1;
       log_struct_ptr->data[1] = data1_q16;
       log_struct_ptr->data[2] = raw_data;
      }
     else
     {
       log_struct_ptr->num_data_types = 5;
       log_struct_ptr->data[0] = data1;
       log_struct_ptr->data[1] = data1_q16;
       log_struct_ptr->data[2] = data2;
       log_struct_ptr->data[3] = data2_q16;
       log_struct_ptr->data[4] = raw_data;
     }

     /* Commit log (also frees up the log packet memory) */
     err_code = sns_logpkt_commit(log_pkt_type,
                                  log_struct_ptr);
   }
   else
   {
       DDF_MSG_1(HIGH, "AL3320 Log Data - logpkt_malloc failed with err: %d", err_code);
       dd_ptr->dropped_logs++;
   }
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_get_mlux

DESCRIPTION   set attrib
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_get_mlux(sns_dd_als_state_t* state)
{
    uint32_t als_lux;
    uint32_t max_lux = 0x7fff; // 32767 depend on user case
    float als_gain_ratio[] = {AL3320_ALS_GAIN_33_RATION, AL3320_ALS_GAIN_08_RATION, AL3320_ALS_GAIN_02_RATION, AL3320_ALS_GAIN_01_RATION};

    als_lux = (float)state->chip.data.als.data * als_gain_ratio[state->chip.setup.als.gain];
    als_lux = als_lux > max_lux?max_lux:als_lux;

    state->chip.data.als.lux = als_lux;
    state->chip.data.als.mlux = als_lux * 1000;
    state->chip.data.als.lux_ok = true;
    DDF_MSG_2(MED, "al3320_get_mlux() end, als lux %d, mlux %d", state->chip.data.als.lux, state->chip.data.als.mlux);

    return SNS_DDF_SUCCESS;
}

/*===========================================================================
FUNCTION      sns_dd_al3320_als_auto_gain

DESCRIPTION   als auto gain
===========================================================================*/
bool sns_dd_al3320_als_auto_gain(sns_dd_als_state_t* state, uint16_t alsdata)
{
    sns_ddf_status_e status = SNS_DDF_SUCCESS;
    uint8_t byteVal;


    status = sns_dd_als_al3320_read_byte(state->port_handle, AL3320_ALS_CON, &byteVal);
    if (status != SNS_DDF_SUCCESS) {
        DDF_MSG_1(HIGH, "al3320_als_auto_gain() read ALS config err: %d", status);
        state->als_auto_gain_trig_time = 0;
        return false;
    }
    state->chip.setup.als.gain = (byteVal & AL3320_ALS_GAIN_MASK) >> AL3320_ALS_GAIN_SHIFT;

    if ((alsdata >= SNS_ALS_AUTO_GAIN_HTHRES) && (state->chip.setup.als.gain > 0x00)) {
        DDF_MSG_2(MED, "al3320_als_auto_gain() als_data: %d > auto_auto_gain_Hthres: %d", alsdata, SNS_ALS_AUTO_GAIN_HTHRES);
        state->chip.setup.als.gain = state->chip.setup.als.gain - 1;
        byteVal = state->chip.setup.als.gain << AL3320_ALS_GAIN_SHIFT;
        status = sns_dd_als_al3320_write_byte(state->port_handle, AL3320_ALS_CON, &byteVal);
        if ( status != SNS_DDF_SUCCESS ) {
            DDF_MSG_1(MED, "al3320_als_auto_gain() write ALS config err: %d", state);
            state->als_auto_gain_trig_time = 0;
            return false;
        }
        state->als_auto_gain_trig_time = sns_ddf_get_timestamp();
    } else if ((alsdata <= SNS_ALS_AUTO_GAIN_LTHRES) && (state->chip.setup.als.gain < 0x03)) {
        DDF_MSG_2(MED, "al3320_als_auto_gain() als_data %d < auto_gain_Lthres: %d", alsdata, SNS_ALS_AUTO_GAIN_LTHRES);
        state->chip.setup.als.gain = state->chip.setup.als.gain + 1;        
        byteVal = state->chip.setup.als.gain << AL3320_ALS_GAIN_SHIFT;
        status = sns_dd_als_al3320_write_byte(state->port_handle, AL3320_ALS_CON, &byteVal);
        if ( status != SNS_DDF_SUCCESS ) {
            DDF_MSG_1(MED, "al3320_als_auto_gain() write ALS config err: %d", state);
            state->als_auto_gain_trig_time = 0;
            return false;
        }
        state->als_auto_gain_trig_time = sns_ddf_get_timestamp();
    } else {
        state->als_auto_gain_trig_time = 0;
        return false;
    }

    DDF_MSG_1(MED, "al3320_als_auto_gain() end, als auto gain triggered, next_gain: %d", state->chip.setup.als.gain);

    return true;
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_get_als_data

DESCRIPTION   get als data from device for al3320 only
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_get_als_data(sns_dd_als_state_t* state)
{
    sns_ddf_status_e      status;
    uword_u alsdata;    


    status = sns_dd_als_al3320_read_word(state->port_handle,
                            AL3320_ALS_LDATA_L,
                               &alsdata);
    if (status != SNS_DDF_SUCCESS)
    {
        DDF_MSG_1(HIGH, "al3320_get_als_data() read als data err: %d", state);
        return status;
    }

#if AL3320_ALS_AUTO_GAIN_EN
    state->als_auto_gain_trig = sns_dd_al3320_als_auto_gain(state, alsdata.word);
#endif

    state->chip.data.als.data = alsdata.word;
    DDF_MSG_3(MED, "al3320_get_als_data() end, alsdata: %d, als gain: %d, als_auto_gain_trig: %d", alsdata.word, state->chip.setup.als.gain, state->als_auto_gain_trig);

    return status;
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_sensor_samples

DESCRIPTION   als report handle
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_sensor_samples(sns_dd_als_state_t    *state,
                                            sns_ddf_sensor_data_s    *data_ptr,
                                            sns_ddf_memhandler_s     *memhandler,
                                             sns_ddf_sensor_sample_s *sample_data)
{
    uint32_t num_samples = 2;

    if(memhandler != NULL)
    {
        data_ptr->samples = sns_ddf_memhandler_malloc_ex(
                                                      memhandler,
                                                      num_samples * sizeof(sns_ddf_sensor_sample_s), state->smgr_handle);
        if(data_ptr->samples == NULL)
        {
            DDF_MSG_0(MED, "al3320_sensor_samples() samples malloc err");
            return SNS_DDF_ENOMEM;
        }
    }
    else if(sample_data != NULL)
    {
        data_ptr->samples = sample_data;
    }
    else
    {
        DDF_MSG_0(MED, "al3320_sensor_samples() samples data err");
        return SNS_DDF_EINVALID_PARAM;
    }

    data_ptr->samples[0].sample = FX_CONV_Q16(state->chip.data.als.lux, 0);
    data_ptr->samples[0].status = SNS_DDF_SUCCESS;
    data_ptr->samples[1].sample = state->chip.data.als.data;
    data_ptr->samples[1].status = SNS_DDF_SUCCESS;
    data_ptr->num_samples       = num_samples;

    sns_dd_als_al3320_log(state, SNS_DDF_SENSOR_AMBIENT,
                              state->chip.data.als.lux, data_ptr->samples[0].sample, 0, 0, (uint32_t) state->chip.data.als.data);

    return SNS_DDF_SUCCESS;
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_set_als_thresh

DESCRIPTION   set als threshold
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_set_als_thresh
(
  sns_dd_als_state_t* state
)
{
    uint32_t alsdata, t_low, t_high;
    uword_u wordVal;

    sns_ddf_status_e status = SNS_DDF_SUCCESS;

    alsdata   = state->chip.data.als.data;
    DDF_MSG_1(MED, "al3320_set_als_thresh() start, alsdata %d",  alsdata );

    if(alsdata < SNS_ALS_INT_CNT_L) alsdata = SNS_ALS_INT_CNT_L;
    t_low = alsdata - SNS_ALS_INT_CNT_L;   /*set to 10 as default*/

    if(alsdata > SNS_ALS_INT_CNT_H ) alsdata = SNS_ALS_INT_CNT_H;
    t_high = alsdata + SNS_ALS_INT_CNT_L;   /*set to 10 as default*/

    wordVal.word = t_low;
    DDF_MSG_1(MED, "al3320_set_als_thresh() als t low wordval.word %d",  wordVal.word);
    status = sns_dd_als_al3320_write_word(state->port_handle, AL3320_ALS_LTHL, &wordVal);
    if ( status != SNS_DDF_SUCCESS )
    {
        DDF_MSG_1(HIGH, "al3320_set_als_thresh() ALS thresh low set error %d", wordVal.word);
        return status;
    }

    wordVal.word = t_high;
    DDF_MSG_1(MED, "al3320_set_als_thresh() als t high wordval.word %d",  wordVal.word);
    status = sns_dd_als_al3320_write_word(state->port_handle, AL3320_ALS_HTHL, &wordVal);
    if ( status != SNS_DDF_SUCCESS )
    {
        DDF_MSG_1(HIGH, "al3320_set_als_thresh() ALS Thresh set high error %d", wordVal.word);
        return status;
    }

    DDF_MSG_0(MED, "al3320_set_als_thresh() end");

    return status;
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_read_status_reg

DESCRIPTION   get data and report
===========================================================================*/
void sns_dd_als_al3320_read_status_reg(sns_dd_als_state_t* state, sns_ddf_time_t timestamp, AL3320_SIG_TYPE sig_type)
{
    uint8_t  statusReg;
    sns_ddf_status_e      status;


#if AL3320_ALS_AUTO_GAIN_EN
    if (state->als_auto_gain_trig && ((timestamp - state->als_auto_gain_trig_time) < state->als_auto_gain_delay_time)) {
        DDF_MSG_2(HIGH, "al3320_read_status_reg() als auto gain waiting time triggered return, pass time: %d, als_auto_gain_usdelay: %d",
                                        (timestamp - state->als_auto_gain_trig_time), state->als_auto_gain_delay_time);
        return;
    }
#endif

    sns_dd_als_al3320_read_byte(state->port_handle,
                      AL3320_FLAG_STATUS,
                      &statusReg);
    DDF_MSG_1(MED, "al3320_read_status_reg() start, status_reg is: 0x%02x" , statusReg);

    if (statusReg & AL3320_STA_ALS_INTR)
    {
        /* Read ALS data */
        status = sns_dd_als_al3320_get_als_data(state);
        if (status != SNS_DDF_SUCCESS) {
            DDF_MSG_1(MED, "al3320_read_status_reg() read als data err: %d", status);
            return;
        }

#if AL3320_ALS_AUTO_GAIN_EN
        if (state->als_auto_gain_trig) {
            DDF_MSG_0(HIGH, "al3320_read_status_reg() als auto gain triggered and returns");
            return;
        }
#endif
        sns_dd_als_al3320_set_als_thresh(state); /* set als threshold to be +/- 20% of current value for next interrupt. */
        /* compute LUX */
        sns_dd_als_al3320_get_mlux(state);
        DDF_MSG_1(MED, "al3320_read_status_reg() ALS lux: %d", state->chip.data.als.lux);

        state->s_data->sensor    = SNS_DDF_SENSOR_AMBIENT;
        state->s_data->status    = SNS_DDF_SUCCESS;
        state->s_data->timestamp = timestamp;
        sns_dd_als_al3320_sensor_samples(state, state->s_data, NULL, state->samples);
        sns_ddf_smgr_notify_data(state->smgr_handle, state->s_data, 1);
    }

    DDF_MSG_0(MED, "al3320_read_status_reg() end");
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_handle_irq

DESCRIPTION   irq handle
===========================================================================*/
void sns_dd_als_al3320_handle_irq
(
  sns_ddf_handle_t  handle,
  uint32_t          gpio_num,
  sns_ddf_time_t    timestamp
)
{
    sns_dd_als_state_t* state = (sns_dd_als_state_t*)handle;
    uint8_t byteVal, i = 0;
    sns_ddf_status_e status;


    DDF_MSG_0(MED, "al3320_handle_irq() start");
    sns_dd_als_al3320_read_status_reg(state, timestamp, IRQ_TYPE);

    /* Clear Interrupt Flag */
    byteVal = SNS_CLR_ALL_INT;
    while ((status = sns_dd_als_al3320_write_byte(state->port_handle, AL3320_FLAG_STATUS, &byteVal)) != SNS_DDF_SUCCESS) {
        if (i == 10) {
            DDF_MSG_1(MED, "al3320_handle_irq() write int_status register err: %d", status);
            break;
        }
        i++;
        sns_ddf_delay(5000);
    }
    /*****/
}

/*===========================================================================
FUNCTION      sns_dd_als_al3320_get_data

DESCRIPTION   get als and prox data
===========================================================================*/
sns_ddf_status_e sns_dd_als_al3320_get_data
(
  sns_ddf_handle_t        dd_handle,
  sns_ddf_sensor_e        sensors[],
  uint32_t                num_sensors,
  sns_ddf_memhandler_s*   memhandler,
  sns_ddf_sensor_data_s*  data[] /* ignored by this async driver */
)
{
    uint8_t i;
    sns_ddf_time_t        timestamp;
    sns_ddf_status_e      status;
    sns_ddf_sensor_data_s *data_ptr;
    sns_dd_als_state_t *state = dd_handle;    
        
    DDF_MSG_0(HIGH, "al3320_get_data() start");

    timestamp = sns_ddf_get_timestamp();

#if AL3320_ALS_AUTO_GAIN_EN
    if (state->als_auto_gain_trig && ((timestamp - state->als_auto_gain_trig_time) < state->als_auto_gain_delay_time)) {
        DDF_MSG_2(HIGH, "als_al3320_get_data() als auto gain waiting time triggered return, passed time: %d, als_auto_gain_usdelay: %d",
                                     (timestamp - state->als_auto_gain_trig_time), state->als_auto_gain_delay_time);
        return SNS_DDF_EFAIL;
    }
#endif

    data_ptr  = sns_ddf_memhandler_malloc_ex(
                                         memhandler,
                                         num_sensors * sizeof(sns_ddf_sensor_data_s), state->smgr_handle);
    if(data_ptr == NULL)
    {
        DDF_MSG_1(HIGH, "al3320_get_data() malloc error, Size: %d", (num_sensors * sizeof(sns_ddf_sensor_data_s)));
        return SNS_DDF_ENOMEM;
    }
    *data = data_ptr;

    for(i = 0; i < num_sensors; i++)
    {
        data_ptr[i].sensor    = sensors[i];
        data_ptr[i].status    = SNS_DDF_SUCCESS;
        data_ptr[i].timestamp = timestamp;

        switch(sensors[i])
        {
        case SNS_DDF_SENSOR_AMBIENT:
            if((status = sns_dd_als_al3320_get_als_data(state))  != SNS_DDF_SUCCESS)
            {
                DDF_MSG_1(HIGH, "al3320_get_data() get als err: %d", status);
                return status;
            }

#if AL3320_ALS_AUTO_GAIN_EN
            if (state->als_auto_gain_trig) {
                DDF_MSG_0(HIGH, "al3320_get_data() als auto gain triggered and returns directly");
                return SNS_DDF_EFAIL;
            }
#endif

            if((status = sns_dd_als_al3320_get_mlux(state)) != SNS_DDF_SUCCESS)
            {
                DDF_MSG_1(HIGH, "al3320_get_data() get mlux err: %d", status);
                return status;
            }
            sns_dd_als_al3320_sensor_samples(state, &data_ptr[i], memhandler, NULL);
            break;

        default:
            DDF_MSG_0(MED, "al3320_get_data() not ALS called");
         break;
        }
    }

    return SNS_DDF_SUCCESS;
}


sns_ddf_driver_if_s AL3320_DD_IF =
{
  .init = &sns_dd_als_al3320_init,
  .get_data = &sns_dd_als_al3320_get_data,
  .set_attrib = &sns_dd_als_al3320_set_attrib,
  .get_attrib = &sns_dd_als_al3320_get_attrib,
  .handle_timer = NULL,
  .handle_irq = &sns_dd_als_al3320_handle_irq,
  .reset = &sns_dd_als_al3320_reset,
  .run_test = &sns_dd_als_al3320_run_test,
  .enable_sched_data = &sns_dd_als_al3320_enable_sched_data,
  .probe = &sns_dd_als_al3320_probe
};

